# coding: utf-8
#
"""
uiautomator2 cli

Usage:
    u2cli install <ip> <url> [--server=<server>]
    u2cli runyaml [--debug] [--step] <filename>

Options:
    -h --help            show this help message
    -v --version         show version
    -s --server=<server>    atx-server url, eg: http://10.0.0.1:8000
    --serial=<serial>    device serial number
    --debug              set loglevel to DEBUG

"""
# u2cli install <url> [--serial=<serial>]

import time
import requests
import re
import six
import sys
import humanize
import hashlib
from docopt import docopt

import uiautomator2
from uiautomator2.cli import runyaml

urllib = six.moves.urllib


def reformat_addr(addr):
    if not re.match(r"^https?://", addr):
        addr = "http://" + addr
    u = urllib.parse.urlparse(addr)
    return u.scheme + "://" + u.netloc


class HTTPError(Exception):
    pass


def raise_for_status(r):
    if r.status_code != 200:
        raise HTTPError(r.text)


def show_pushing_progress(ret, start_time):
    """
    Args:
        ret: json message from URL(/install/:id)
    """
    if ret is None:
        print("No progress")
        return
    total = ret.get('totalSize', 0)
    if total == 0:
        print("Total size 0")
        return
    copied = ret.get('copiedSize', 0)
    total_size = humanize.naturalsize(total, gnu=True)
    copied_size = humanize.naturalsize(copied, gnu=True)
    speed = humanize.naturalsize(
        (copied / (time.time() - start_time)), gnu=True)
    progress = 100.0 * float(copied) / float(total)
    print("Pushing {:.1f}% {} / {} [{}B/s]".format(progress, copied_size,
                                                   total_size, speed))


class Installer(object):
    def __init__(self, device_url, server_url):
        self._device_url = reformat_addr(device_url)
        self._server_url = server_url
        self._devinfo = None

    @property
    def devinfo(self):
        if self._devinfo:
            return self._devinfo
        self._devinfo = requests.get(self._device_url + "/info").json()
        return self._devinfo

    @property
    def serial(self):
        return self.devinfo['serial']

    def vivo_install(self, apk_url):
        print("Vivo detected, open u2 watchers")
        u = uiautomator2.connect_wifi(self._device_url)
        u.watcher("AUTO_INSTALL").when(
            textMatches="好|安装", className="android.widget.Button").click()
        u.watchers.watched = True
        self.pm_install(apk_url)

    def oppo_install(self, apk_url):
        m = hashlib.md5()
        m.update(apk_url.encode('utf-8'))
        key = m.hexdigest()[8:16]
        dst = "/sdcard/atx-" + key + ".apk"  # 安装包会在安装完后自动删除
        d = uiautomator2.connect(self._device_url)
        print(d.info, dst)
        d.push_url(apk_url, dst)
        # d.push("/Users/codeskyblue/workdir/atxcrawler/apks/ApiDemos-debug.apk", dst) # For debug
        with d.session("com.coloros.filemanager") as s:
            s(text=u"所有文件").click()
            s(className="android.widget.ListView").scroll.to(textContains=key)
            s(textContains=key).click()

            btn_done = d(className="android.widget.Button", text=u"完成")
            while not btn_done.exists:
                s(text="继续安装旧版本").click_exists()
                s(text="无视风险安装").click_exists()
                s(text="重新安装").click_exists()
                # 自动清除安装包和残留
                if s(resourceId=
                     "com.android.packageinstaller:id/install_confirm_panel"
                     ).exists:
                    # 通过偏移点击<安装>
                    s(resourceId=
                      "com.android.packageinstaller:id/bottom_button_layout"
                      ).click(offset=(0.75, 0.5))
            btn_done.click()

    def pm_install(self, apk_url):
        install_url = self._install_url()
        r = requests.post(install_url, data={'url': apk_url})
        print(r.text)
        raise_for_status(r)
        id = r.text.strip()
        u = urllib.parse.urlparse(install_url)
        query_url = u.scheme + "://" + u.netloc + "/install/" + id
        try:
            self._wait_finish(query_url)
        except KeyboardInterrupt:
            print("Catch Interrupt signal, cancel download")
            r = requests.delete(query_url)
            print(r.text)

    def install(self, apk_url):
        brand = self.devinfo.get('brand', '').lower()
        if brand == 'vivo':
            self.vivo_install(apk_url)
        elif brand == 'oppo':
            self.oppo_install(apk_url)
        else:
            self.pm_install(apk_url)

    def _provider_install_url(self):
        if not self._server_url:
            return
        server_url = reformat_addr(self._server_url)
        dinfo = requests.get(
            server_url + "/devices/" + self.devinfo['udid'] + "/info").json()
        provider = dinfo.get('provider')
        if not provider:
            return None
        return 'http://{}:{}/install/{}'.format(
            provider['ip'], provider['port'], dinfo['serial'])

    def _device_install_url(self):
        return self._device_url + "/install"

    def _install_url(self):
        purl = self._provider_install_url()
        if purl:
            print("Use provider install service:", purl)
            return purl
        iurl = self._device_install_url()
        print("Use device install service:", iurl)
        return iurl

    def _wait_finish(self, query_url):
        start = time.time()
        while True:
            time.sleep(1)
            r = requests.get(query_url)
            raise_for_status(r)
            ret = r.json()
            if not ret:
                print("wait progress info")
                continue
            # raise when error found
            err = ret.get('error')
            if err:
                print("Unexpected error:", err)
                raise RuntimeError(err)
            # message is also status
            status = ret.get('message', '')
            if status == 'finished':
                print("Success installed")
                return True
            elif status == 'pushing':
                show_pushing_progress(ret, start)
            elif status == 'downloading':  # for old style
                show_pushing_progress(ret.get('progress', {}), start)
            elif status == 'installing':
                print("Installing ..")
            elif status == 'success installed':
                print("Installed")
                return
            elif status and status.startswith("err:"):
                raise RuntimeError(status, ret)
            elif status and 'error' in status:
                raise RuntimeError(status, ret)
            else:
                print(ret)


__commands = {}


def register_command(func, name=None, args=()):
    name = name or func.__name__
    __commands[name] = (func, args)


def __cmd_install(ip, server, apk_url):
    print("CMD:", ip, server, apk_url)
    device_url = ip + ":7912"
    ins = Installer(device_url, server)
    ins.install(apk_url)


def __cmd_runyaml(debug, onlystep, filename):
    try:
        import yaml
    except ImportError:
        sys.exit("you need to install pyaml")
    runyaml.main(filename, debug, onlystep)


def main():
    args = docopt(__doc__, version='u2cli 1.0')
    print(args)
    register_command(__cmd_install, 'install', ('<ip>', '--server', '<url>'))
    register_command(__cmd_runyaml, "runyaml", ('--debug', '--step', '<filename>'))

    for cmdname, cmdopts in __commands.items():
        if args[cmdname]:
            func, argnames = cmdopts
            cmdargs = [args[argname] for argname in argnames]
            func(*cmdargs)

        # return
        # print("InstallURL", install_url)
        # install_url = get_install_url(args['<ip>'], args['--server'])
        # # https://gohttp.nie.netease.com/tools/apks/qrcodescan-2.6.0-green.apk'
        # install_apk(install_url, apk_url)



#!/usr/bin/env python
# coding: utf-8
#
"""
packageName: com.netease.cloudmusic  # Optional
activity: .MainActivity # Optional
launch: true # Optional, default true
close: true # Optional, close app when finished
plugins:
  ocr: $ocr-api-url
  perf:
    interval: 1
    debug: true
    filename: perf.csv
steps:
- q: ~hell
- q: ^hello w
- q: =hello world
- q: hello world
- q: foo
  timeout: 10 # seconds
- text: the input text
- code: |
    time.sleep(10) # only "d, time" can be used
watchers: # click when show up
- q: ~跳过
  timeout: 0 # default
"""

import argparse
import time
import re
import logging

import uiautomator2 as u2
import logzero
from logzero import logger


class JSONRunner(object):
    __alias = {
        'q': 'query',
    }

    def __init__(self, cnf):
        self._cnf = cnf
        self._title = cnf.get('title')
        self._pkg_name = cnf.get('packageName')
        self._activity = cnf.get('activity')
        self._launch = cnf.get('launch', True)
        self._close = cnf.get('close', True)
        self._clear = cnf.get('clear')
        self._steps = cnf.get('steps')
        self._watchers = cnf.get('watchers', [])
        self._udid = cnf.get('udid')
        self._plugins = cnf.get('plugins', {})
        d = self._d = u2.connect(self._udid)

        self.session = None

        # plugins
        if 'ocr' in self._plugins:
            logger.info("load plugin: ocr")
            import uiautomator2.ext.ocr as ocr
            ocr.API = self._plugins['ocr']
            logger.info("Use ocr plugin: %s", ocr.API)
            u2.plugin_register('ocr', ocr.OCR)
        if self._pkg_name and 'perf' in self._plugins:
            logger.info("load plugin: perf")
            import uiautomator2.ext.perf as perf
            u2.plugin_register('perf', perf.Perf, self._pkg_name)
            perfcnf = self._plugins['perf']
            d.ext_perf.interval = perfcnf.get('interval', 1.0)
            d.ext_perf.debug = perfcnf.get('debug', True)
            d.ext_perf.csv_output = perfcnf.get('filename', 'perf.csv')

    def _find_xpath(self, xpath, hierarchy):
        el = self.session.xpath(xpath, hierarchy)
        if el.exists:
            return el.wait().center()

    def _find_text(self, text, hierarchy):
        elements = self.session.xpath(
            '//*[re:match(name(), "\.(TextView|Button|ImageView)$")]',
            hierarchy).all()
        if text.startswith('~') or text.startswith('^'):  # Regexp
            pattern = re.compile(text[1:] if text.startswith('~') else text)
            for el in elements:
                if el.text and pattern.match(el.text):
                    logger.debug("find match: %s, %s", el.text, el.attrib)
                    return el.center()
        elif text.startswith("="):
            text = text[1:]

        for el in elements:
            if el.text == text:
                logger.debug("find exactly match text: %s", el.text)
                return el.center()

    def _oper_input(self, text):
        logger.info("input text: %s", text)
        self.session.set_fastinput_ime(True)
        self.session.send_keys(text)
        self.session.press("enter")

    def _run_watchers(self, hierarchy):
        for kwargs in self._watchers:
            kwargs['timeout'] = kwargs.get('timeout', 0)
            if 'q' in kwargs:
                kwargs['query'] = kwargs.pop('q')
            if self._run_onestep(hierarchy, **kwargs):
                logger.info("trigger watcher: %s", kwargs)

    def _run_onestep(self,
                     hierarchy,
                     action='click',
                     query=None,
                     text=None,
                     code=None,
                     ocr=None,
                     timeout=10):
        """
        Returns:
            bool: if step handled
        """
        if text:
            self._oper_input(text)
            return True

        if ocr:
            self._d.ext_ocr(ocr).click(timeout=timeout)
            return True

        # find element and click
        if query:
            pos = None
            if query.startswith("/"):
                pos = self._find_xpath(query, hierarchy)
            else:
                pos = self._find_text(query, hierarchy)

            if pos is None:
                if action == 'assertNotExists':
                    return True
                return False

            if action == 'click':
                logger.info("click: %s", pos)
                self.session.click(*pos)
                return True
            elif action == 'assertExists':
                return True

        if code:
            logger.info("exec: |\n%s", code)
            exec(code, {'d': self.session, 'time': time})
            return True

        # raise NotImplementedError("only support click action")

    def _handle_step(self, **kwargs):
        retry_cnt = 0
        timeout = kwargs.pop('timeout', 10)
        deadline = time.time() + timeout

        # alias
        if 'q' in kwargs:
            kwargs['query'] = kwargs.pop('q')

        while retry_cnt == 0 or time.time() < deadline:
            retry_cnt += 1
            hierarchy = self.session.dump_hierarchy()
            self._run_watchers(hierarchy)  # 处理弹框
            if self._run_onestep(hierarchy, **kwargs):
                break
            logger.debug("process %s, retry %d", kwargs, retry_cnt)
            time.sleep(.5)
        else:
            raise RuntimeError("element not found: %s", kwargs)

    def prepare_session(self):
        d = self._d
        if not self._launch or not self._pkg_name:
            self.session = self._d.session(self._pkg_name, attach=True)
            return

        if self._clear:
            d.app_clear(self._pkg_name)

        s = None
        if self._activity:
            d.app_start(self._pkg_name, self._activity, stop=True)
            s = d.session(self._pkg_name, attach=True)
        else:
            s = d.session(self._pkg_name)

        self.session = s

    def run(self):
        logger.info("test begins: %s", self._title)
        logger.info("launch app: %s", self._pkg_name)

        self.prepare_session()
        if 'perf' in self._plugins:
            self._d.ext_perf.start()

        try:
            for step in self._steps:
                logger.info("==> %s", step)
                self._handle_step(**step)
            logger.info("Finished")
        finally:
            if 'perf' in self._plugins:
                self._d.ext_perf.stop()
            if self._launch and self._pkg_name and self._close:
                self._d.app_stop(self._pkg_name)


def main(filename, debug=False, onlystep=False):
    if not debug:
        logzero.loglevel(logging.INFO)

    import yaml
    with open(filename, 'rb') as f:
        cnf = yaml.load(f)
        if onlystep:
            cnf['launch'] = False
        tc = JSONRunner(cnf)
        tc.run()


# if __name__ == '__main__':
#     parser = argparse.ArgumentParser()
#     parser.add_argument(
#         '-d', '--debug', help='set loglevel to debug', action='store_true')
#     parser.add_argument('--step', help='run only step', action='store_true')
#     parser.add_argument('yamlfile')
#     args = parser.parse_args()

#     main(args.yamlfile, args.debug, onlystep=args.step)



# coding: utf-8
#

from __future__ import print_function
from __future__ import absolute_import

import fire
import os
import logging
import subprocess
import shutil
import tarfile
import hashlib
import re
import time
import socket
import progress.bar
from contextlib import closing

import humanize
import requests

import uiautomator2 as u2
from uiautomator2 import adbutils
from uiautomator2.version import __apk_version__, __atx_agent_version__


def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        '%(asctime)s - %(filename)s:%(lineno)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


log = get_logger('uiautomator2')
appdir = os.path.join(os.path.expanduser("~"), '.uiautomator2')
log.debug("use cache directory: %s", appdir)

GITHUB_BASEURL = "https://github-mirror.open.netease.com/openatx"


class DownloadBar(progress.bar.Bar):
    message = "Downloading"
    suffix = '%(current_size)s / %(total_size)s'

    @property
    def total_size(self):
        return humanize.naturalsize(self.max, gnu=True)

    @property
    def current_size(self):
        return humanize.naturalsize(self.index, gnu=True)


def cache_download(url, filename=None):
    """ return downloaded filepath """
    # check cache
    if not filename:
        filename = os.path.basename(url)
    storepath = os.path.join(appdir,
                             hashlib.sha224(url.encode()).hexdigest(),
                             filename)
    storedir = os.path.dirname(storepath)
    if not os.path.isdir(storedir):
        os.makedirs(storedir)
    if os.path.exists(storepath) and os.path.getsize(storepath) > 0:
        log.debug("file '%s' cached before", filename)
        return storepath
    # download from url
    r = requests.get(url, stream=True)
    log.debug("download from %s", url)
    if r.status_code != 200:
        raise Exception("status code", r.status_code)
    file_size = int(r.headers.get("Content-Length"))
    bar = DownloadBar(filename, max=file_size)
    with open(storepath + '.tmp', 'wb') as f:
        chunk_length = 16 * 1024
        while 1:
            buf = r.raw.read(chunk_length)
            if not buf:
                break
            f.write(buf)
            bar.next(len(buf))
        bar.finish()
    shutil.move(storepath + '.tmp', storepath)
    return storepath


class Installer(adbutils.Adb):
    def __init__(self, serial=None):
        super(Installer, self).__init__(serial)
        self.sdk = self.getprop('ro.build.version.sdk')
        self.abi = self.getprop('ro.product.cpu.abi')
        self.pre = self.getprop('ro.build.version.preview_sdk')
        self.arch = self.getprop('ro.arch')
        self.server_addr = None

    def get_executable_dir(self):
        return "/data/local/tmp"
        # Hot patch for some phone which not contains stat command
        # dirs = ['/data/local/tmp', '/data/data/com.android.shell']
        # for dirname in dirs:
        #     testpath = "%s/%s" % (dirname, 'permtest')
        #     self.shell('touch', testpath, raise_error=False)
        #     self.shell('chmod', '+x', testpath, raise_error=False)
        #     content = self.shell('stat', '-c%A', testpath, raise_error=False)
        #     # log.debug('stat returns:', content)
        #     if -1 != content.find('x'):
        #         return dirname
        # return "/data/local/tmp"
        # raise EnvironmentError("Can't find an executable directory on device")

    def install_minicap(self):
        if self.arch == 'x86':
            log.info("skip install minicap on emulator")
            return
        sdk = self.sdk
        if self.pre and self.pre != "0":
            sdk = sdk + self.pre
        base_url = GITHUB_BASEURL + \
            "/stf-binaries/raw/master/node_modules/minicap-prebuilt/prebuilt/"
        log.debug("install minicap.so")
        url = base_url + self.abi + "/lib/android-" + sdk + "/minicap.so"
        path = cache_download(url)
        exedir = self.get_executable_dir()
        minicapdst = "%s/%s" % (exedir, 'minicap.so')
        self.push(path, minicapdst)
        log.info("install minicap")
        url = base_url + self.abi + "/bin/minicap"
        path = cache_download(url)
        self.push(path, exedir + "/minicap", 0o755)

    def install_minitouch(self):
        """ Need test """
        log.info("install minitouch")
        url = ''.join([
            GITHUB_BASEURL + "/stf-binaries",
            "/raw/master/node_modules/minitouch-prebuilt/prebuilt/",
            self.abi + "/bin/minitouch"
        ])
        path = cache_download(url)
        exedir = self.get_executable_dir()
        self.push(path, exedir + "/minitouch", 0o755)

    def download_uiautomator_apk(self, apk_version):
        app_url = GITHUB_BASEURL + \
            '/android-uiautomator-server/releases/download/%s/app-uiautomator.apk' % apk_version
        app_test_url = GITHUB_BASEURL + \
            '/android-uiautomator-server/releases/download/%s/app-uiautomator-test.apk' % apk_version
        log.info("app-uiautomator.apk(%s) downloading ...", apk_version)
        path = cache_download(app_url)

        log.info("app-uiautomator-test.apk downloading ...")
        pathtest = cache_download(app_test_url)
        return (path, pathtest)

    def install_uiautomator_apk(self, apk_version, reinstall=False):
        pkg_info = self.package_info('com.github.uiautomator')
        test_pkg_info = self.package_info('com.github.uiautomator.test')
        # For test_pkg_info has no versionName or versionCode
        # Just check if the com.github.uiautomator.test apk is installed
        if not reinstall and pkg_info and pkg_info['version_name'] == apk_version and test_pkg_info:
            log.info("apk(%s) already installed, skip", apk_version)
            return
        if pkg_info or test_pkg_info:
            log.debug("uninstall old apks")
            self.uninstall('com.github.uiautomator')
            self.uninstall('com.github.uiautomator.test')

        (path, pathtest) = self.download_uiautomator_apk(apk_version)
        self.install(path)
        log.debug("app-uiautomator.apk installed")

        self.install(pathtest)
        log.debug("app-uiautomator-test.apk installed")

    def check_apk_installed(self, apk_version):
        """ in OPPO device, if you check immediatelly, package_info will return None """
        pkg_info = self.package_info("com.github.uiautomator")
        if not pkg_info:
            raise EnvironmentError(
                "package com.github.uiautomator not installed")
        if pkg_info['version_name'] != apk_version:
            raise EnvironmentError(
                "package com.github.uiautomator version expect \"%s\" got \"%s\""
                % (apk_version, pkg_info['version_name']))
        # test apk
        pkg_test_info = self.package_info("com.github.uiautomator.test")
        if not pkg_test_info:
            raise EnvironmentError(
                "package com.github.uiautomator.test not installed")

    def check_agent_installed(self, agent_version):
        lport = self.forward_port(7912)
        log.debug("forward device(port:7912) -> %d", lport)
        try:
            r = requests.get("http://127.0.0.1:%d/version" % lport, timeout=5)
            return r.text.strip() == agent_version
        except:
            return False

    def install_atx_agent(self, agent_version, reinstall=False):
        version_output = self.shell(
            '/data/local/tmp/atx-agent', '-v', raise_error=False).strip()
        m = re.search(r"\d+\.\d+\.\d+", version_output)
        current_agent_version = m.group(0) if m else None
        if current_agent_version == agent_version:
            log.info("atx-agent(%s) already installed, skip", agent_version)
            return
        if current_agent_version == 'dev' and not reinstall:
            log.warn("atx-agent develop version, skip")
            return
        if current_agent_version:
            log.info("atx-agent(%s) need to update", current_agent_version)
        files = {
            'armeabi-v7a': 'atx-agent_{v}_linux_armv7.tar.gz',
            'arm64-v8a': 'atx-agent_{v}_linux_armv7.tar.gz',
            'armeabi': 'atx-agent_{v}_linux_armv6.tar.gz',
            'x86': 'atx-agent_{v}_linux_386.tar.gz',
        }
        log.info("atx-agent(%s) is installing, please be patient",
                 agent_version)
        abis = self.shell('getprop',
                          'ro.product.cpu.abilist').strip() or self.abi
        name = None
        for abi in abis.split(','):
            name = files.get(abi)
            if name:
                break
        if not name:
            raise Exception(
                "arch(%s) need to be supported yet, please report an issue in github"
                % abis)
        url = GITHUB_BASEURL + '/atx-agent/releases/download/%s/%s' % (
            agent_version, name.format(v=agent_version))
        log.debug("download atx-agent(%s) from github releases", agent_version)
        path = cache_download(url)
        tar = tarfile.open(path, 'r:gz')
        bin_path = os.path.join(os.path.dirname(path), 'atx-agent')
        tar.extract('atx-agent', os.path.dirname(bin_path))
        self.push(bin_path, '/data/local/tmp/atx-agent', 0o755)
        log.debug("atx-agent installed")

    def launch_and_check(self):
        log.info("launch atx-agent daemon")
        exedir = self.get_executable_dir()
        exefile = "%s/%s" % (exedir, 'atx-agent')
        args = ['TMPDIR=/sdcard', exefile, '-d']
        if self.server_addr:
            args.append('-t')
            args.append(self.server_addr)
        output = self.shell(*args)
        lport = self.forward_port(7912)
        log.debug("forward device(port:7912) -> %d", lport)
        time.sleep(.5)
        cnt = 0
        while cnt < 3:
            try:
                r = requests.get(
                    'http://localhost:%d/version' % lport, timeout=10)
                log.debug("atx-agent version: %s", r.text)
                # todo finish the retry logic
                print("atx-agent output:", output.strip())
                # open uiautomator2 github URL
                self.shell("am", "start", "-a", "android.intent.action.VIEW",
                           "-d", "https://github.com/openatx/uiautomator2")
                log.info("success")
                break
            except (requests.exceptions.ConnectionError,
                    requests.exceptions.ReadTimeout):
                time.sleep(.5)
                cnt += 1
        else:
            log.error(
                "Failure, unable to get result from http://localhost:%d/version"
            )


class MyFire(object):
    def init(self,
             server=None,
             apk_version=__apk_version__,
             agent_version=__atx_agent_version__,
             verbose=False,
             reinstall=False,
             ignore_apk_check=False,
             proxy=None,
             serial=None,
             mirror=False):
        if verbose:
            log.setLevel(logging.DEBUG)
        if server:
            log.info("atx-server addr %s", server)
        if mirror:
            global GITHUB_BASEURL
            GITHUB_BASEURL = "https://github-mirror.open.netease.com/openatx"

        if proxy:
            os.environ['HTTP_PROXY'] = proxy
            os.environ['HTTPS_PROXY'] = proxy

        if not serial:
            valid_serials = [
                sn for sn, _ in adbutils.Adb().devices(states=['device'])
            ]
            # output = subprocess.check_output(['adb', 'devices'])
            # pattern = re.compile(
            #     r'(?P<serial>[^\s]+)\t(?P<status>device|offline)')
            # matches = pattern.findall(output.decode())
            # valid_serials = [m[0] for m in matches if m[1] == 'device']
            if len(valid_serials) == 0:
                log.warning("No avaliable android devices detected.")
                return
            log.info("Detect pluged devices: %s", valid_serials)
            for serial in valid_serials:
                self._init_with_serial(serial, server, apk_version,
                                       agent_version, reinstall,
                                       ignore_apk_check)
        else:
            self._init_with_serial(serial, server, apk_version, agent_version,
                                   reinstall, ignore_apk_check)

    def _init_with_serial(self, serial, server, apk_version, agent_version,
                          reinstall, ignore_apk_check):
        log.info("Device(%s) initialing ...", serial)
        ins = Installer(serial)
        ins.server_addr = server
        ins.install_minicap()
        ins.install_minitouch()
        ins.install_uiautomator_apk(apk_version, reinstall)

        if ins.check_agent_installed(agent_version):
            # TODO: should also check atx-server addr
            log.info("atx-agent is already running, force stop")
            ins.shell("/data/local/tmp/atx-agent", "--stop", raise_error=False)
            ins.shell("rm", "/sdcard/atx-agent.pid", raise_error=False)
            ins.shell("rm", "/sdcard/atx-agent.log.old", raise_error=False)
        else:
            ins.install_atx_agent(agent_version, reinstall)

        if not ignore_apk_check:
            ins.check_apk_installed(apk_version)
        ins.launch_and_check()

    def update_apk(self, ip):
        """ update com.github.uiautomator apk remotely """
        u = u2.connect(ip)
        apk_version = __apk_version__
        app_url = GITHUB_BASEURL + \
            '/android-uiautomator-server/releases/download/%s/app-uiautomator.apk' % apk_version
        app_test_url = GITHUB_BASEURL + \
            '/android-uiautomator-server/releases/download/%s/app-uiautomator-test.apk' % apk_version
        u.app_install(app_url)
        u.app_install(app_test_url)

    def clear_cache(self):
        log.info("clear cache dir: %s", appdir)
        shutil.rmtree(appdir, ignore_errors=True)

    def cleanup(self):
        raise NotImplementedError()

    def install(self, arg1, arg2=None):
        """
        Example:
            install "http://some-host.apk"
            install "$serial" "http://some-host.apk"
        """
        if arg2 is None:
            device_ip, apk_url = None, arg1
        else:
            device_ip, apk_url = arg1, arg2
        u = u2.connect(device_ip)
        pkg_name = u.app_install(apk_url)
        print("Installed", pkg_name)

    def unlock(self, device_ip=None):
        u = u2.connect(device_ip)
        u.unlock()

    def app_stop_all(self, device_ip=None):
        u = u2.connect(device_ip)
        u.app_stop_all()

    def uninstall_all(self, device_ip=None):
        u = u2.connect(device_ip)
        u.app_uninstall_all(verbose=True)

    def identify(self, device_ip=None, theme='black'):
        u = u2.connect(device_ip)
        u.open_identify(theme)

    def screenshot(self, device_ip, filename):
        u = u2.connect(device_ip)
        u.screenshot(filename)

    def healthcheck(self, device_ip):
        u = u2.connect(device_ip)
        u.healthcheck()


def main():
    fire.Fire(MyFire)


if __name__ == '__main__':
    main()



# coding: utf-8
#

from __future__ import absolute_import, print_function

import threading
import re
import time
import datetime
import csv
import sys
import atexit
from collections import namedtuple

_MEM_PATTERN = re.compile(r'TOTAL[:\s]+(\d+)')
# acct_tag_hex is a socket tag
# cnt_set==0 are for background data
# cnt_set==1 are for foreground data
_NetStats = namedtuple(
    "NetStats",
    """idx iface acct_tag_hex uid_tag_int cnt_set rx_bytes rx_packets
    tx_bytes tx_packets rx_tcp_bytes rx_tcp_packets rx_udp_bytes rx_udp_packets rx_other_bytes rx_other_packets
    tx_tcp_bytes tx_tcp_packets tx_udp_bytes tx_udp_packets tx_other_bytes tx_other_packets"""
    .split())


class Perf(object):
    def __init__(self, d, package_name=None):
        self.d = d
        self.package_name = package_name
        self.csv_output = "perf.csv"
        self.debug = False
        self.interval = 1.0
        self._th = None
        self._event = threading.Event()
        self._condition = threading.Condition()
        self._data = {}

    def memory(self):
        """ PSS(KB) """
        output = self.d.shell(['dumpsys', 'meminfo', self.package_name]).output
        m = _MEM_PATTERN.search(output)
        if m:
            return int(m.group(1))
        return 0

    def _cpu_rawdata_collect(self, pid):
        first_line = self.d.shell(['cat', '/proc/stat']).output.splitlines()[0]
        assert first_line.startswith('cpu ')
        # ds: user, nice, system, idle, iowait, irq, softirq, stealstolen, guest, guest_nice
        ds = list(map(int, first_line.split()[1:]))
        total_cpu = sum(ds)
        idle = ds[3]

        proc_stat = self.d.shell(
            ['cat', '/proc/%d/stat' % pid]).output.split(') ')[1].split()
        utime = int(proc_stat[11])
        stime = int(proc_stat[12])
        return (total_cpu, idle, utime + stime)

    def cpu(self, pid):
        """ CPU

        Refs:
        - http://man7.org/linux/man-pages/man5/proc.5.html
        - [安卓性能测试之cpu占用率统计方法总结](https://www.jianshu.com/p/6bf564f7cdf0)
        """
        store_key = 'cpu-%d' % pid
        # first time jiffies
        if store_key in self._data:
            tjiff1, idle1, pjiff1 = self._data[store_key]
        else:
            tjiff1, idle1, pjiff1 = self._cpu_rawdata_collect(pid)
            time.sleep(.3)

        # second time jiffies
        self._data[
            store_key] = tjiff2, idle2, pjiff2 = self._cpu_rawdata_collect(pid)

        # calculate
        pcpu = 100.0 * (pjiff2 - pjiff1) / (tjiff2 - tjiff1)  # process cpu
        scpu = 100.0 * ((tjiff2 - idle2) -
                        (tjiff1 - idle1)) / (tjiff2 - tjiff1)  # system cpu
        return round(pcpu, 1), round(scpu, 1)

        # Retrive cpu from top
        # for line in d.shell(["top", "-n", "1"]).output.splitlines():
        #     vs = line.split()
        #     if len(vs) > 5 and vs[-1] == self.package_name:
        #         if vs[4].endswith('%'):
        #             print("Miss:", float(vs[4][:-1]) - pcpu)
        #             return float(vs[4][:-1]), 0.0
        # return (0.0, 0.0)

    def netstat(self, pid):
        """
        Returns:
            (rx_bytes, tx_bytes)
        """
        m = re.search(r'^Uid:\s+(\d+)',
                      self.d.shell(['cat', '/proc/%d/status' % pid]).output,
                      re.M)
        if not m:
            return (0, 0)
        uid = m.group(1)
        lines = self.d.shell(
            ['cat', '/proc/net/xt_qtaguid/stats']).output.splitlines()

        rx, tx = 0, 0
        for line in lines:
            vs = line.split()
            if len(vs) != 21:
                continue
            v = _NetStats(*vs)
            if v.uid_tag_int != uid:
                continue
            if v.iface != 'wlan0':
                continue
            # FIXME(ssx): tcp and udp data will support when some one needed
            rx += int(v.rx_bytes)
            tx += int(v.tx_bytes)

        store_key = 'netstat-%s' % uid
        drx, dtx = 0, 0
        if store_key in self._data:
            last_rx, last_tx = self._data[store_key]
            drx, dtx = rx - last_rx, tx - last_tx
        self._data[store_key] = (rx, tx)
        return drx, dtx

    def collect(self):
        pid = self.d._pidof_app(self.package_name)
        if pid is None:
            return
        pss = self.memory()
        cpu, scpu = self.cpu(pid)
        rx_bytes, tx_bytes = self.netstat(pid)
        timestr = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        return {
            'time': timestr,
            'pss': round(pss / 1024.0, 2),
            'cpu': cpu,
            'systemCpu': scpu,
            'rxBytes': rx_bytes,
            'txBytes': tx_bytes,
        }

    def continue_collect(self, f):
        try:
            headers = ['time', 'pss', 'cpu', 'systemCpu', 'rxBytes', 'txBytes']
            fcsv = csv.writer(f)
            fcsv.writerow(headers)
            update_time = time.time()
            while not self._event.isSet():
                perfdata = self.collect()
                if self.debug:
                    print("DEBUG:", perfdata)
                if not perfdata:
                    print("perf package is not alive:", self.package_name)
                    time.sleep(1)
                    continue
                fcsv.writerow([perfdata[k] for k in headers])
                wait_seconds = max(0,
                                   self.interval - (time.time() - update_time))
                time.sleep(wait_seconds)
                update_time = time.time()
            f.close()
        finally:
            self._condition.acquire()
            self._th = None
            self._condition.notify()
            self._condition.release()

    def start(self):
        if sys.version_info.major < 3:
            f = open(self.csv_output, "wb")
        else:
            f = open(self.csv_output, "w", newline='\n')

        def defer_close():
            if not f.closed:
                f.close()

        atexit.register(defer_close)

        if self._th:
            raise RuntimeError("perf is already running")
        if not self.package_name:
            raise EnvironmentError("package_name need to be set")
        self._data.clear()
        self._event = threading.Event()
        self._condition = threading.Condition()
        self._th = threading.Thread(target=self.continue_collect, args=(f, ))
        self._th.daemon = True
        self._th.start()

    def stop(self):
        self._event.set()
        self._condition.acquire()
        self._condition.wait(timeout=2)
        self._condition.release()
        if self.debug:
            print("DEBUG: perf collect stopped")


if __name__ == '__main__':
    import uiautomator2 as u2
    pkgname = "com.tencent.tmgp.sgame"
    pkgname = "com.netease.cloudmusic"
    u2.plugin_register('perf', Perf, pkgname)

    d = u2.connect("10.242.62.224")
    print(d.current_app())
    # print(d.ext_perf.netstat(5350))
    # d.app_start(pkgname)
    d.ext_perf.start()
    d.ext_perf.debug = True
    try:
        time.sleep(500)
    except KeyboardInterrupt:
        d.ext_perf.stop()
        print("threading stopped")